
#include "ble_sfst_tag.h"
#include <string.h>
#include "nordic_common.h"
#include "ble_srv_common.h"
#include "app_util.h"




/* Function for handling the BLE Connect event.
 * Sets the p_tag connection handle to be used later by the client
 * PARAMS
 *    p_tag      - a pointer to the ble_tag_t structure that contains our service details (p_tag)
 *    p_ble_evt  - a pointer to the event details structure received from the BLE stack
 */
static void on_connect(ble_tag_t *p_tag,  ble_evt_t *p_ble_evt)
{
    p_tag->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}




/* Function for handling the BLE Disconnect event.
 * Resets the p_tag connection handle to invalid 0xFFFF after a disconnect
 * PARAMS
 *    p_tag      - a pointer to the ble_tag_t structure that contains our service details (p_tag)
 *    p_ble_evt  - a pointer to the event details structure not required for this implementation
 */
static void on_disconnect(ble_tag_t  *p_tag,  ble_evt_t *p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt); //not used in this function set to void
    p_tag->conn_handle = BLE_CONN_HANDLE_INVALID; //reset the handle
}




/* Function for handling the Write message received from the Client
 * PARAMS
 *    p_tag      - a pointer to the ble_tag_t structure that contains our service details (p_tag)
 *    p_ble_evt  - a pointer to the event details structure received from the BLE stack
 */
static void on_write(ble_tag_t *p_tag,  ble_evt_t *p_ble_evt)
{
    ble_gatts_evt_write_t *p_evt_write = &p_ble_evt->evt.gatts_evt.params.write; //copy the write event details 
    
    if ((p_evt_write->handle == p_tag->led_char_handles.value_handle) &&  //make sure the write is for our LED characteristic
        (p_evt_write->len == 1) &&																				//verify we received 1 byte of data to write
        (p_tag->led_write_handler != NULL))																//verify the led_write_handler() has been initialized
    {
        p_tag->led_write_handler(p_tag, p_evt_write->data[0]);	//call our led_write_handler() function and send the data to be written
    }
}




/* Main event handler for parsing the events received from the BLE Stack for this service
 * PARAMS
 *    p_tag      - a pointer to the ble_tag_t structure that contains our service details (p_tag)
 *    p_ble_evt  - a pointer to the event details structure received from the BLE stack
 */
void ble_tag_on_ble_evt(ble_tag_t *p_tag, ble_evt_t *p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_tag, p_ble_evt); //call the on_connect() service function (see above)
            break;
            
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_tag, p_ble_evt);
            break;
            
        case BLE_GATTS_EVT_WRITE:
            on_write(p_tag, p_ble_evt);
            break;
            
        default:
            // No implementation needed.
            break;
    }
}




/* Function for adding the LED characteristic.
 * The LED state needs to be readable and writable without any notification 
 */
static uint32_t led_char_add(ble_tag_t *p_tag, const ble_tag_init_t *p_tag_init)
{
    //Define the structures that will hold our Characteristic's information
    ble_uuid_t              ble_uuid;					//holds the UUID for this characteristic  
	  ble_gatts_char_md_t     char_md;          //Characteristic Metadata
    ble_gatts_attr_md_t     attr_md;          //LED state attribute metadata
	  ble_gatts_attr_t        attr_char_value;  //LED state attribute value

	
		//Each Characteristic has a UUID. Assign it here
    ble_uuid.type = p_tag->uuid_type;
    ble_uuid.uuid = TAG_UUID_LED_CHAR;


		/*** LED CHARACTERISTIC METADATA ***/
	  //The characteristic has metadata which describes it's configuration
    memset(&char_md, 0, sizeof(char_md));		//iniitalize characteristic metatadata to 0
    char_md.char_props.read   = 1;					//LED status can be read
    char_md.char_props.write  = 1;					//LED status can be written
    char_md.p_char_user_desc  = NULL;				//Description string not used
    char_md.p_char_pf         = NULL;				//BLE presentation format not used
    char_md.p_user_desc_md    = NULL;				//Description metadata not needed
    char_md.p_cccd_md         = NULL;		    //Notify not used, so no need for the Client CCD
    char_md.p_sccd_md         = NULL;				//Server CCD not used
    
   
	 	/*** LED STATE ATTRIBUTE ***/
		//The LED characteristic has a single attribute for the LED state value
		//First, define the metadata for the LED State Attribute		
    memset(&attr_md, 0, sizeof(attr_md));                //iniitalize attribute metatadata to 0
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);  //LED status can be read
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm); //LED status can be written
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;           //Place the attribute metadata structure on the stack so it can be shared
    attr_md.rd_auth    = 0;                              //these will be set by the Application at runtime
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;
    
    //Next, configure the structure that holds the Attribute value and link it to the Attribute metadata defined above
		memset(&attr_char_value, 0, sizeof(attr_char_value));
    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;             //link to the Attribute Metadata
    attr_char_value.init_len     = sizeof(uint8_t);
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = sizeof(uint8_t);
    attr_char_value.p_value      = NULL;
    
		
		//Add the LED characteristic to the softdevice database so it can be accessed via BLE
    return sd_ble_gatts_characteristic_add(p_tag->service_handle, &char_md, &attr_char_value, &p_tag->led_char_handles);
}
//////END led_char_add()



/* Function for adding the Button characteristic.
 * The Button characteristic Notifies the client on State change and allows the client to read the current button state
 * Since we are using Notifications of state changes, we need to add the CCCD structure so the client can turn on
 * notification when it connects to the device
 */
static uint32_t button_char_add(ble_tag_t *p_tag, const ble_tag_init_t *p_tag_init)
{
    //Define the structures that will hold our Characteristic's information
    ble_uuid_t              ble_uuid;					//holds the UUID for this characteristic  
	  ble_gatts_char_md_t     char_md;          //Characteristic Metadata
	  ble_gatts_attr_md_t     cccd_md;          //CCCD used to enable notification of button state changes. We use an attribute metadata structure for this
    ble_gatts_attr_md_t     attr_md;          //button state attribute metadata
	  ble_gatts_attr_t        attr_char_value;  //button state attribute value
    

		//Each Characteristic has a UUID. Assign it here
    ble_uuid.type = p_tag->uuid_type;
    ble_uuid.uuid = TAG_UUID_BUTTON_CHAR;
		
	
		/*** BUTTON CHARACTERISTIC METADATA ***/
	  //The characteristic has metadata which describes it's configuration. We configure it in two steps
	  //Step 1. Configure the Client Configuration Descriptor for the button characteristic metadata
    memset(&cccd_md, 0, sizeof(cccd_md));
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);   //Macro configures the CCCD to allow reading the CCCD over BLE
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);  //Macro configures the CCCD to allow writing the CCCD over BLE
    cccd_md.vloc = BLE_GATTS_VLOC_STACK;                  //Place the CCCD metadata structure on the stack so it can be shared

    //Step 2. Configure the rest of the button characteristic metadata and link it to the CCCD structure from Step 1
    memset(&char_md, 0, sizeof(char_md));
    char_md.char_props.read   = 1;					//Permit reading of button state
    char_md.char_props.notify = 1;					//Permit notifications of button state changes
    char_md.p_char_user_desc  = NULL;				//Description string not used
    char_md.p_char_pf         = NULL;				//BLE presentation format not used
    char_md.p_user_desc_md    = NULL;				//Description metadata not needed
    char_md.p_cccd_md         = &cccd_md;		//pointer to the Client Characteristic Config structure we populated above
    char_md.p_sccd_md         = NULL;				//Server CCD not used
    
		
		/*** BUTTON BUTTON STATE ATTRIBUTE ***/
		//The button characteristic has a single attribute for the button state value
		//First, define the metadata for the Button State Attribute		
		memset(&attr_md, 0, sizeof(attr_md));
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);        //Button state can be read
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);  //Button state can NOT be written
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;								 //Place the attribute metadata structure on the stack so it can be shared by the appication and the softdevice
    attr_md.rd_auth    = 0;																		 //Authorization will be requested from the Application, so set to 0 for now
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;
    
		//Configure the structure that holds the Button State value and link it to the Attribute metadata defined above
    memset(&attr_char_value, 0, sizeof(attr_char_value));
    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;  //associate the Attribute value with the Attribute metadata structure above
    attr_char_value.init_len     = sizeof(uint8_t);
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = sizeof(uint8_t);
    attr_char_value.p_value      = NULL;
    
		
		//Add the button characteristic to the softdevice database so it can be accessed via BLE
    return sd_ble_gatts_characteristic_add(p_tag->service_handle, &char_md, &attr_char_value, &p_tag->button_char_handles);
}
//////END button_char_add()



uint32_t ble_tag_init(ble_tag_t *p_tag, const ble_tag_init_t *p_tag_init)
{
    uint32_t   err_code;
    ble_uuid_t ble_uuid;

    // Initialize service structure
    p_tag->conn_handle       = BLE_CONN_HANDLE_INVALID;
    p_tag->led_write_handler = p_tag_init->led_write_handler;
    
    // Add service
    ble_uuid128_t base_uuid = {TAG_UUID_BASE};
    err_code = sd_ble_uuid_vs_add(&base_uuid, &p_tag->uuid_type);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    
    ble_uuid.type = p_tag->uuid_type;
    ble_uuid.uuid = TAG_UUID_SERVICE;

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_tag->service_handle);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    
    err_code = button_char_add(p_tag, p_tag_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    
    err_code = led_char_add(p_tag, p_tag_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    
    return NRF_SUCCESS;
}



//This function sends a BLE Notification of the Button Change Event
uint32_t ble_tag_on_button_change(ble_tag_t *p_tag, uint8_t button_state)
{
    ble_gatts_hvx_params_t params;
    uint16_t len = sizeof(button_state);
    
    memset(&params, 0, sizeof(params));
    params.type = BLE_GATT_HVX_NOTIFICATION;
    params.handle = p_tag->button_char_handles.value_handle;
    params.p_data = &button_state;
    params.p_len = &len;
    
    return sd_ble_gatts_hvx(p_tag->conn_handle, &params);
}
